@@ -31,12 +31,15 @@ from reprounzip.unpackers.common import COMPAT_OK, COMPAT_MAYBE, \
     make_unique_name, shell_escape, select_installer, busybox_url, sudo_url, \
     FileUploader, FileDownloader, get_runs, add_environment_options, \
     fixup_environment, interruptible_call, metadata_read, metadata_write, \
-    metadata_initial_iofiles, metadata_update_run
+    metadata_initial_iofiles, metadata_update_run, parse_ports
 from reprounzip.unpackers.common.x11 import X11Handler, LocalForwarder
 from reprounzip.utils import unicode_, iteritems, stderr, join_root, \
     download_file
 
 
+logger = logging.getLogger('reprounzip.docker')
+
+
 # How this all works:
 #  - setup/create just copies file to the target directory and writes the
 #    Dockerfile
@@ -66,10 +69,10 @@ def select_image(runs):
                                                  architecture))
 
     if architecture == 'i686':
-        logging.info("Wanted architecture was i686, but we'll use x86_64 with "
-                     "Docker")
+        logger.info("Wanted architecture was i686, but we'll use x86_64 with "
+                    "Docker")
     elif architecture != 'x86_64':
-        logging.error("Error: unsupported architecture %s", architecture)
+        logger.error("Error: unsupported architecture %s", architecture)
         sys.exit(1)
 
     def find_distribution(parameter, distribution, version):
@@ -82,8 +85,8 @@ def select_image(runs):
                 if result is not None:
                     return result
         distrib = images[default]
-        logging.warning("Unsupported distribution '%s', using %s",
-                        distribution, default)
+        logger.warning("Unsupported distribution '%s', using %s",
+                       distribution, default)
         return find_version(distrib, None)
 
     def find_version(distrib, version):
@@ -93,8 +96,8 @@ def select_image(runs):
                     return image['distribution'], image['image']
         image = distrib['default']
         if version is not None:
-            logging.warning("Using %s instead of '%s'",
-                            image['name'], version)
+            logger.warning("Using %s instead of '%s'",
+                           image['name'], version)
         return image['distribution'], image['image']
 
     return find_distribution(get_parameter('docker_images'),
@@ -117,7 +120,7 @@ def docker_setup(args):
     docker_setup_create(args)
     try:
         docker_setup_build(args)
-    except:
+    except BaseException:
         Path(args.target[0]).rmtree(ignore_errors=True)
         raise
 
@@ -128,7 +131,7 @@ def docker_setup_create(args):
     pack = Path(args.pack[0])
     target = Path(args.target[0])
     if target.exists():
-        logging.critical("Target directory exists")
+        logger.critical("Target directory exists")
         sys.exit(1)
 
     signals.pre_setup(target=target, pack=pack)
@@ -153,15 +156,15 @@ def docker_setup_create(args):
                 target_distribution = None
         else:
             target_distribution, base_image = select_image(runs)
-        logging.info("Using base image %s", base_image)
-        logging.debug("Distribution: %s", target_distribution or "unknown")
+        logger.info("Using base image %s", base_image)
+        logger.debug("Distribution: %s", target_distribution or "unknown")
 
         rpz_pack.copy_data_tar(target / 'data.tgz')
 
         arch = runs[0]['architecture']
 
         # Writes Dockerfile
-        logging.info("Writing %s...", target / 'Dockerfile')
+        logger.info("Writing %s...", target / 'Dockerfile')
         with (target / 'Dockerfile').open('w', encoding='utf-8',
                                           newline='\n') as fp:
             fp.write('FROM %s\n\n' % base_image)
@@ -196,9 +199,9 @@ def docker_setup_create(args):
                     installer = select_installer(pack, runs,
                                                  target_distribution)
                 except CantFindInstaller as e:
-                    logging.error("Need to install %d packages but couldn't "
-                                  "select a package installer: %s",
-                                  len(packages), e)
+                    logger.error("Need to install %d packages but couldn't "
+                                 "select a package installer: %s",
+                                 len(packages), e)
                     sys.exit(1)
                 # Updates package sources
                 update_script = installer.update_script()
@@ -206,8 +209,8 @@ def docker_setup_create(args):
                     fp.write('    %s && \\\n' % update_script)
                 # Installs necessary packages
                 fp.write('    %s && \\\n' % installer.install_script(packages))
-                logging.info("Dockerfile will install the %d software "
-                             "packages that were not packed", len(packages))
+                logger.info("Dockerfile will install the %d software "
+                            "packages that were not packed", len(packages))
             else:
                 record_usage(docker_install_pkgs=False)
 
@@ -215,7 +218,7 @@ def docker_setup_create(args):
             paths = set()
             pathlist = []
             # Add intermediate directories, and check for existence in the tar
-            logging.info("Generating file list...")
+            logger.info("Generating file list...")
             missing_files = chain.from_iterable(pkg.files
                                                 for pkg in missing_packages)
             data_files = rpz_pack.data_filenames()
@@ -235,7 +238,7 @@ def docker_setup_create(args):
                     if path in data_files:
                         pathlist.append(path)
                     else:
-                        logging.info("Missing file %s", path)
+                        logger.info("Missing file %s", path)
             rpz_pack.close()
             # FIXME : for some reason we need reversed() here, I'm not sure why
             # Need to read more of tar's docs.
@@ -266,27 +269,41 @@ def docker_setup_build(args):
     target = Path(args.target[0])
     unpacked_info = read_dict(target)
     if 'initial_image' in unpacked_info:
-        logging.critical("Image already built")
+        logger.critical("Image already built")
         sys.exit(1)
 
-    image = make_unique_name(b'reprounzip_image_')
+    if args.image_name:
+        image = args.image_name[0]
+    else:
+        image = make_unique_name(b'reprounzip_image_')
 
-    logging.info("Calling 'docker build'...")
+    logger.info("Calling 'docker build'...")
     try:
-        retcode = subprocess.call(['docker', 'build', '-t'] +
+        retcode = subprocess.call(args.docker_cmd.split() + ['build', '-t'] +
                                   args.docker_option + [image, '.'],
                                   cwd=target.path)
     except OSError:
-        logging.critical("docker executable not found")
+        logger.critical("docker executable not found")
         sys.exit(1)
     else:
         if retcode != 0:
-            logging.critical("docker build failed with code %d", retcode)
+            logger.critical("docker build failed with code %d", retcode)
             sys.exit(1)
-    logging.info("Initial image created: %s", image.decode('ascii'))
+    logger.info("Initial image created: %s", image.decode('ascii'))
 
     unpacked_info['initial_image'] = image
     unpacked_info['current_image'] = image
+    if 'DOCKER_MACHINE_NAME' in os.environ:
+        unpacked_info['docker_host'] = {
+            'type': 'docker-machine',
+            'name': os.environ['DOCKER_MACHINE_NAME']}
+    elif 'DOCKER_HOST' in os.environ:
+        unpacked_info['docker_host'] = {
+            'type': 'custom',
+            'env': dict((k, v)
+                        for k, v in iteritems(os.environ)
+                        if k.startswith('DOCKER_'))}
+
     write_dict(target, unpacked_info)
 
 
@@ -300,20 +317,20 @@ def docker_reset(args):
     target = Path(args.target[0])
     unpacked_info = read_dict(target)
     if 'initial_image' not in unpacked_info:
-        logging.critical("Image doesn't exist yet, have you run setup/build?")
+        logger.critical("Image doesn't exist yet, have you run setup/build?")
         sys.exit(1)
     image = unpacked_info['current_image']
     initial = unpacked_info['initial_image']
 
     if image == initial:
-        logging.warning("Image is already in the initial state, nothing to "
-                        "reset")
+        logger.warning("Image is already in the initial state, nothing to "
+                       "reset")
     else:
-        logging.info("Removing image %s", image.decode('ascii'))
-        retcode = subprocess.call(['docker', 'rmi', image])
+        logger.info("Removing image %s", image.decode('ascii'))
+        retcode = subprocess.call(args.docker_cmd.split() + ['rmi', image])
         if retcode != 0:
-            logging.warning("Can't remove previous image, docker returned %d",
-                            retcode)
+            logger.warning("Can't remove previous image, docker returned %d",
+                           retcode)
         unpacked_info['current_image'] = initial
         write_dict(target, unpacked_info)
 
@@ -387,7 +404,7 @@ def docker_run(args):
 
     # Sanity check
     if args.detach and args.x11:
-        logging.critical("Error: Can't use X11 forwarding if you're detaching")
+        logger.critical("Error: Can't use X11 forwarding if you're detaching")
         raise UsageError
 
     # Loads config
@@ -399,9 +416,9 @@ def docker_run(args):
     # Get current image name
     if 'current_image' in unpacked_info:
         image = unpacked_info['current_image']
-        logging.debug("Running from image %s", image.decode('ascii'))
+        logger.debug("Running from image %s", image.decode('ascii'))
     else:
-        logging.critical("Image doesn't exist yet, have you run setup/build?")
+        logger.critical("Image doesn't exist yet, have you run setup/build?")
         sys.exit(1)
 
     # Name of new container
@@ -412,6 +429,12 @@ def docker_run(args):
 
     hostname = runs[selected_runs[0]].get('hostname', 'reprounzip')
 
+    # Port forwarding
+    port_options = []
+    for port_host, port_container, proto in parse_ports(args.expose_port):
+        port_options.extend(['-p',
+                             '%s:%s%s' % (port_host, port_container, proto)])
+
     # X11 handler
     if args.x11:
         local_ip = get_local_addr()
@@ -433,7 +456,7 @@ def docker_run(args):
                 ssh_cmdline = ' '.join(
                     '-R*:%(p)d:127.0.0.1:%(p)d' % {'p': port}
                     for port, connector in x11.port_forward)
-                logging.warning(
+                logger.warning(
                     "You requested X11 forwarding but the Docker container "
                     "appears to be running remotely. It is probable that it "
                     "won't be able to connect to the local display. Creating "
@@ -476,61 +499,69 @@ def docker_run(args):
         forwarders.append(LocalForwarder(connector, port))
 
     if args.detach:
-        logging.info("Start container %s (detached)",
-                     container.decode('ascii'))
-        retcode = interruptible_call(['docker', 'run', b'--name=' + container,
+        logger.info("Start container %s (detached)",
+                    container.decode('ascii'))
+        retcode = interruptible_call(args.docker_cmd.split() +
+                                     ['run', b'--name=' + container,
                                       '-h', hostname,
                                       '-d', '-t'] +
+                                     port_options +
                                      args.docker_option +
                                      [image, '/busybox', 'sh', '-c', cmds])
         if retcode != 0:
-            logging.critical("docker run failed with code %d", retcode)
+            logger.critical("docker run failed with code %d", retcode)
             subprocess.call(['docker', 'rm', '-f', container])
             sys.exit(1)
         return
 
     # Run command in container
-    logging.info("Starting container %s", container.decode('ascii'))
-    retcode = interruptible_call(['docker', 'run', b'--name=' + container,
+    logger.info("Starting container %s", container.decode('ascii'))
+    retcode = interruptible_call(args.docker_cmd.split() +
+                                 ['run', b'--name=' + container,
                                   '-h', hostname,
                                   '-i', '-t'] +
+                                 port_options +
                                  args.docker_option +
-                                 [image, '/busybox', 'sh', '-c', cmds])
-    if retcode != 0:
-        logging.critical("docker run failed with code %d", retcode)
-        subprocess.call(['docker', 'rm', '-f', container])
-        sys.exit(1)
+                                 [image, '/busybox', 'sh', '-c', cmds],
+                                 request_tty=True)
 
     # Get exit status from "docker inspect"
-    out = subprocess.check_output(['docker', 'inspect', container])
+    try:
+        out = subprocess.check_output(args.docker_cmd.split() +
+                                      ['inspect', container])
+    except subprocess.CalledProcessError:
+        logger.critical("docker run failed with code %d", retcode)
+        subprocess.call(['docker', 'rm', '-f', container])
+        sys.exit(1)
     outjson = json.loads(out.decode('ascii'))
     if (outjson[0]["State"]["Running"] is not False or
             outjson[0]["State"]["Paused"] is not False):
-        logging.error("Invalid container state after execution:\n%s",
-                      json.dumps(outjson[0]["State"]))
+        logger.error("Invalid container state after execution:\n%s",
+                     json.dumps(outjson[0]["State"]))
     retcode = outjson[0]["State"]["ExitCode"]
     stderr.write("\n*** Command finished, status: %d\n" % retcode)
 
     # Commit to create new image
     new_image = make_unique_name(b'reprounzip_image_')
-    logging.info("Committing container %s to image %s",
-                 container.decode('ascii'), new_image.decode('ascii'))
-    subprocess.check_call(['docker', 'commit', container, new_image])
+    logger.info("Committing container %s to image %s",
+                container.decode('ascii'), new_image.decode('ascii'))
+    subprocess.check_call(args.docker_cmd.split() +
+                          ['commit', container, new_image])
 
     # Update image name
     unpacked_info['current_image'] = new_image
     write_dict(target, unpacked_info)
 
     # Remove the container
-    logging.info("Destroying container %s", container.decode('ascii'))
-    retcode = subprocess.call(['docker', 'rm', container])
+    logger.info("Destroying container %s", container.decode('ascii'))
+    retcode = subprocess.call(args.docker_cmd.split() + ['rm', container])
     if retcode != 0:
-        logging.error("Error deleting container %s", container.decode('ascii'))
+        logger.error("Error deleting container %s", container.decode('ascii'))
 
     # Untag previous image, unless it is the initial_image
     if image != unpacked_info['initial_image']:
-        logging.info("Untagging previous image %s", image.decode('ascii'))
-        subprocess.check_call(['docker', 'rmi', image])
+        logger.info("Untagging previous image %s", image.decode('ascii'))
+        subprocess.check_call(args.docker_cmd.split() + ['rmi', image])
 
     # Update input file status
     metadata_update_run(config, unpacked_info, selected_runs)
@@ -540,8 +571,10 @@ def docker_run(args):
 
 
 class ContainerUploader(FileUploader):
-    def __init__(self, target, input_files, files, unpacked_info):
+    def __init__(self, target, input_files, files, unpacked_info,
+                 docker_cmd='docker'):
         self.unpacked_info = unpacked_info
+        self.docker_cmd = docker_cmd
         FileUploader.__init__(self, target, input_files, files)
 
     def prepare_upload(self, files):
@@ -562,7 +595,7 @@ class ContainerUploader(FileUploader):
             name = stem + ('_%d' % nb).encode('ascii') + ext
         name = Path(name)
         local_path.copyfile(self.build_directory / name)
-        logging.info("Copied file %s to %s", local_path, name)
+        logger.info("Copied file %s to %s", local_path, name)
         self.docker_copy.append((name, input_path))
 
     def finalize(self):
@@ -592,20 +625,22 @@ class ContainerUploader(FileUploader):
             # TODO : restore permissions?
 
         image = make_unique_name(b'reprounzip_image_')
-        retcode = subprocess.call(['docker', 'build', '-t', image, '.'],
+        retcode = subprocess.call(self.docker_cmd +
+                                  ['build', '-t', image, '.'],
                                   cwd=self.build_directory.path)
         if retcode != 0:
-            logging.critical("docker build failed with code %d", retcode)
+            logger.critical("docker build failed with code %d", retcode)
             sys.exit(1)
         else:
-            logging.info("New image created: %s", image.decode('ascii'))
+            logger.info("New image created: %s", image.decode('ascii'))
             if from_image != self.unpacked_info['initial_image']:
-                logging.info("Untagging previous image %s",
-                             from_image.decode('ascii'))
-                retcode = subprocess.call(['docker', 'rmi', from_image])
+                logger.info("Untagging previous image %s",
+                            from_image.decode('ascii'))
+                retcode = subprocess.call(self.docker_cmd +
+                                          ['rmi', from_image])
                 if retcode != 0:
-                    logging.warning("Can't remove previous image, docker "
-                                    "returned %d", retcode)
+                    logger.warning("Can't remove previous image, docker "
+                                   "returned %d", retcode)
             self.unpacked_info['current_image'] = image
             write_dict(self.target, self.unpacked_info)
 
@@ -622,21 +657,24 @@ def docker_upload(args):
     input_files = unpacked_info.setdefault('input_files', {})
 
     try:
-        ContainerUploader(target, input_files, files, unpacked_info)
+        ContainerUploader(target, input_files, files, unpacked_info,
+                          docker_cmd=args.docker_cmd.split())
     finally:
         write_dict(target, unpacked_info)
 
 
 class ContainerDownloader(FileDownloader):
-    def __init__(self, target, files, image, all_=False):
+    def __init__(self, target, files, image, all_=False, docker_cmd='docker'):
         self.image = image
+        self.docker_cmd = docker_cmd
         FileDownloader.__init__(self, target, files, all_=all_)
 
     def prepare_download(self, files):
         # Create a container from the image
         self.container = make_unique_name(b'reprounzip_dl_')
-        logging.info("Creating container %s", self.container.decode('ascii'))
-        subprocess.check_call(['docker', 'create',
+        logger.info("Creating container %s", self.container.decode('ascii'))
+        subprocess.check_call(self.docker_cmd +
+                              ['create',
                                b'--name=' + self.container,
                                self.image])
 
@@ -645,11 +683,12 @@ class ContainerDownloader(FileDownloader):
         # a file name (#4272)
         tmpdir = Path.tempdir(prefix='reprozip_docker_output_')
         try:
-            ret = subprocess.call(['docker', 'cp',
-                                  self.container + b':' + remote_path.path,
-                                  tmpdir.path])
+            ret = subprocess.call(self.docker_cmd +
+                                  ['cp',
+                                   self.container + b':' + remote_path.path,
+                                   tmpdir.path])
             if ret != 0:
-                logging.critical("Can't get output file: %s", remote_path)
+                logger.critical("Can't get output file: %s", remote_path)
                 return False
             (tmpdir / remote_path.name).copyfile(local_path)
         finally:
@@ -657,11 +696,11 @@ class ContainerDownloader(FileDownloader):
         return True
 
     def finalize(self):
-        logging.info("Removing container %s", self.container.decode('ascii'))
-        retcode = subprocess.call(['docker', 'rm', self.container])
+        logger.info("Removing container %s", self.container.decode('ascii'))
+        retcode = subprocess.call(self.docker_cmd + ['rm', self.container])
         if retcode != 0:
-            logging.warning("Can't remove temporary container, docker "
-                            "returned %d", retcode)
+            logger.warning("Can't remove temporary container, docker "
+                           "returned %d", retcode)
 
 
 @target_must_exist
@@ -673,12 +712,13 @@ def docker_download(args):
     unpacked_info = read_dict(target)
 
     if 'current_image' not in unpacked_info:
-        logging.critical("Image doesn't exist yet, have you run setup/build?")
+        logger.critical("Image doesn't exist yet, have you run setup/build?")
         sys.exit(1)
     image = unpacked_info['current_image']
-    logging.debug("Downloading from image %s", image.decode('ascii'))
+    logger.debug("Downloading from image %s", image.decode('ascii'))
 
-    ContainerDownloader(target, files, image, all_=args.all)
+    ContainerDownloader(target, files, image,
+                        all_=args.all, docker_cmd=args.docker_cmd.split())
 
 
 @target_must_exist
@@ -688,7 +728,7 @@ def docker_destroy_docker(args):
     target = Path(args.target[0])
     unpacked_info = read_dict(target)
     if 'initial_image' not in unpacked_info:
-        logging.critical("Image not created")
+        logger.critical("Image not created")
         sys.exit(1)
 
     initial_image = unpacked_info.pop('initial_image')
@@ -696,15 +736,15 @@ def docker_destroy_docker(args):
     if 'current_image' in unpacked_info:
         image = unpacked_info.pop('current_image')
         if image != initial_image:
-            logging.info("Destroying image %s...", image.decode('ascii'))
-            retcode = subprocess.call(['docker', 'rmi', image])
+            logger.info("Destroying image %s...", image.decode('ascii'))
+            retcode = subprocess.call(args.docker_cmd.split() + ['rmi', image])
             if retcode != 0:
-                logging.error("Error deleting image %s", image.decode('ascii'))
+                logger.error("Error deleting image %s", image.decode('ascii'))
 
-    logging.info("Destroying image %s...", initial_image.decode('ascii'))
-    retcode = subprocess.call(['docker', 'rmi', initial_image])
+    logger.info("Destroying image %s...", initial_image.decode('ascii'))
+    retcode = subprocess.call(args.docker_cmd.split() + ['rmi', initial_image])
     if retcode != 0:
-        logging.error("Error deleting image %s", initial_image.decode('ascii'))
+        logger.error("Error deleting image %s", initial_image.decode('ascii'))
 
 
 @target_must_exist
@@ -714,7 +754,7 @@ def docker_destroy_dir(args):
     target = Path(args.target[0])
     read_dict(target)
 
-    logging.info("Removing directory %s...", target)
+    logger.info("Removing directory %s...", target)
     signals.pre_destroy(target=target)
     target.rmtree()
     signals.post_destroy(target=target)
@@ -768,14 +808,17 @@ def setup(parser, **kwargs):
       output_id:filename    extracts the output file to the corresponding local
                             path
     """
+    parser.add_argument('--docker-cmd', action='store', default='docker',
+                        help="Change the Docker command; split on spaces")
+
     subparsers = parser.add_subparsers(title="actions",
                                        metavar='', help=argparse.SUPPRESS)
 
     def add_opt_general(opts):
         opts.add_argument('target', nargs=1, help="Experiment directory")
 
-    # setup/create
-    def add_opt_setup(opts):
+    # Common between setup and setup/create
+    def add_opt_setup_create(opts):
         opts.add_argument('pack', nargs=1, help="Pack to extract")
         opts.add_argument('--base-image', nargs=1, help="Base image to use")
         opts.add_argument('--distribution', nargs=1,
@@ -789,6 +832,14 @@ def setup(parser, **kwargs):
                           default=False, dest='install_pkgs',
                           help=argparse.SUPPRESS)
 
+    # Common between setup and setup/build
+    def add_opt_setup_build(opts):
+        opts.add_argument('--image-name', nargs=1,
+                          help="Name of the image to create (by default, a "
+                               "random name beginning with "
+                               "'reprounzip_image_' is generated)")
+        add_raw_docker_option(opts)
+
     # --docker-option
     def add_raw_docker_option(opts):
         opts.add_argument('--docker-option', action='append',
@@ -797,21 +848,21 @@ def setup(parser, **kwargs):
                                "specified multiple times")
 
     parser_setup_create = subparsers.add_parser('setup/create')
-    add_opt_setup(parser_setup_create)
+    add_opt_setup_create(parser_setup_create)
     add_opt_general(parser_setup_create)
     parser_setup_create.set_defaults(func=docker_setup_create)
 
     # setup/build
     parser_setup_build = subparsers.add_parser('setup/build')
     add_opt_general(parser_setup_build)
-    add_raw_docker_option(parser_setup_build)
+    add_opt_setup_build(parser_setup_build)
     parser_setup_build.set_defaults(func=docker_setup_build)
 
     # setup
     parser_setup = subparsers.add_parser('setup')
-    add_opt_setup(parser_setup)
+    add_opt_setup_create(parser_setup)
+    add_opt_setup_build(parser_setup)
     add_opt_general(parser_setup)
-    add_raw_docker_option(parser_setup)
     parser_setup.set_defaults(func=docker_setup)
 
     # reset
@@ -823,7 +874,7 @@ def setup(parser, **kwargs):
     parser_upload = subparsers.add_parser('upload')
     add_opt_general(parser_upload)
     parser_upload.add_argument('file', nargs=argparse.ZERO_OR_MORE,
-                               help="<path>:<input_file_name")
+                               help="<path>:<input_file_name>")
     parser_upload.set_defaults(func=docker_upload)
 
     # run
@@ -832,6 +883,9 @@ def setup(parser, **kwargs):
     parser_run.add_argument('run', default=None, nargs=argparse.OPTIONAL)
     parser_run.add_argument('--cmdline', nargs=argparse.REMAINDER,
                             help="Command line to run")
+    parser_run.add_argument('--expose-port', '-p', action='append', default=[],
+                            help="Expose a network port, "
+                                 "host[:experiment[/proto]]. Example: 8000:80")
     parser_run.add_argument('--enable-x11', action='store_true', default=False,
                             dest='x11',
                             help="Enable X11 support (needs an X server on "
